#include <SDL2/SDL_events.h>
#include <SDL2/SDL_pixels.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_scancode.h>
#include <SDL2/SDL_stdinc.h>
#include <SDL2/SDL_surface.h>
#include <SDL2/SDL_timer.h>
#include <SDL2/SDL_video.h>
#include <SDL2/SDL_image.h>
#include <immintrin.h>
#include <stdint.h>
#include <stdio.h>
#include <SDL2/SDL.h>
#include <complex.h>
#include <omp.h>
#include <stdlib.h>
#include <x86intrin.h>
#include <argp.h>

#include "mandelbrot.h"
#include "mouse.h"
#include "ui.h"

/*TODO
 *  1. Store iterations instead of colors ?
 *      1.1 memory usage vs performance loss ??? (cant draw in parallel anyway)
 *  2. Look into SDL2 rendering
 *      2.1 texture tendering   DONE
 *      2.2 fix colors ??? 
 *  3. SSE2
 *      3.1 x86intrin.h DONE
*/

int
mandelbrot_avx(Complex c, unsigned max_iterations)
{
    __m256 cr = _mm256_set1_pd(c.real);
    __m256 ci = _mm256_set1_pd(c.imag);

    __m256 zr = cr;
    __m256 zi = ci;

    __m256 threshold = _mm256_set1_pd(4); 

    unsigned int k = 1;

    __m256 mk = _mm256_set1_pd(k);
    __m256 one = _mm256_set1_pd(1);
    while (++k < max_iterations){
        /* Compute z1 from z0 */
        __m256 zr2 = _mm256_mul_pd(zr, zr);
        __m256 zi2 = _mm256_mul_pd(zi, zi);
        __m256 zrzi = _mm256_mul_pd(zr, zi);
        /* zr1 = zr0 * zr0 - zi0 * zi0 + cr */
        /* zi1 = zr0 * zi0 + zr0 * zi0 + ci */
        zr = _mm256_add_pd(_mm256_sub_pd(zr2, zi2), cr);
        zi = _mm256_add_pd(_mm256_add_pd(zrzi, zrzi), ci);

        /* Increment k */
        zr2 = _mm256_mul_pd(zr, zr);
        zi2 = _mm256_mul_pd(zi, zi);
        __m256 mag2 = _mm256_add_pd(zr2, zi2);
        __m256 mask = _mm256_cmp_pd(mag2, threshold, _CMP_LT_OS);
        mk = _mm256_add_pd(_mm256_and_pd(mask, one), mk);

        /* Early bailout? */
        if (_mm256_testz_pd(mask, _mm256_set1_pd(-1)))
            break;
    }
    return k;
}


int 
mandelbrot(Complex c, unsigned int max_iterations) {
    Complex z = {0.0, 0.0};
    unsigned int i;
    
    for (i = 0; i < max_iterations; i++) {
        long double temp_real = z.real * z.real - z.imag * z.imag + c.real;
        long double temp_imag = 2 * z.real * z.imag + c.imag;
        z.real = temp_real;
        z.imag = temp_imag;
        
        if (__builtin_hypot(z.real, z.imag) > 2)
            return i;
    }
    return max_iterations;
} 

typedef struct{
    uint8_t r;
    uint8_t g;
    uint8_t b;
    uint8_t a;
} Color;

typedef struct {
    Color *pointer;
    int size;
}Array;

void
calculate_set(Array *arr, int height, int width, ViewInfo view, short unsigned use_avx, unsigned int max_iterations)
{
    int (*mandelbrot_func)(Complex C, unsigned);

    if (use_avx == 0)
        mandelbrot_func = mandelbrot;
    else
        mandelbrot_func = mandelbrot_avx;

    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            long double real = view.x_min + (x * (view.x_max - view.x_min)) / width;
            long double imag = view.y_min + (y * (view.y_max - view.y_min)) / height;
            Complex c = {real, imag};
            
            unsigned int iterations = mandelbrot_func(c, max_iterations);

            if (iterations == max_iterations){
                arr->pointer[y * width + x].r = 0;
                arr->pointer[y * width + x].g = 0;
                arr->pointer[y * width + x].b = 0;
                arr->pointer[y * width + x].a = 0;
            } else {
                double t = (double)iterations / max_iterations;
                t = 0.5 + 0.5 * tan(log(t + 0.0001) * 3.0);

                arr->pointer[y * width + x].r = 255 * t * 0.5;
                arr->pointer[y * width + x].g = 255 * t * 0.4;
                arr->pointer[y * width + x].b = 255 * t;
                arr->pointer[y * width + x].a = 255;
            }
        }
    }
}

void
calculate_set_omp(Array *arr, int height, int width, ViewInfo view, short unsigned use_avx, unsigned int max_iterations)
{
    int (*mandelbrot_func)(Complex, unsigned);

    if (use_avx == 0)
        mandelbrot_func = mandelbrot;
    else
        mandelbrot_func = mandelbrot_avx;

    #pragma omp parallel for schedule(guided)
    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            long double real = view.x_min + (x * (view.x_max - view.x_min)) / width;
            long double imag = view.y_min + (y * (view.y_max - view.y_min)) / height;
            Complex c = {real, imag};
            
            unsigned int iterations = mandelbrot_func(c, max_iterations);

            if (iterations == max_iterations){
                arr->pointer[y * width + x].r = 0;
                arr->pointer[y * width + x].g = 0;
                arr->pointer[y * width + x].b = 0;
                arr->pointer[y * width + x].a = 0;
            } else {
                double t = (double)iterations / max_iterations;
                t = 0.5 + 0.5 * tan(log(t + 0.0001) * 3.0);

                arr->pointer[y * width + x].r = 255 * t * 0.5;
                arr->pointer[y * width + x].g = 255 * t * 0.4;
                arr->pointer[y * width + x].b = 255 * t;
                arr->pointer[y * width + x].a = 255;
            }
        }
    }
}

void 
update_texture(int width, int height, SDL_Texture *texture, Array *arr) {
    void *pixels;
    int pitch;
    
    if (SDL_LockTexture(texture, NULL, &pixels, &pitch) != 0) {
        fprintf(stderr, "Failed to lock texture: %s\n", SDL_GetError());
        return;
    }
    
    memcpy(pixels, arr->pointer, width* height * sizeof(Color));
    
    SDL_UnlockTexture(texture);
}


struct arguments
{
    char *output_file;
    int width;
    int height;
    unsigned int max_iterations;
    short unsigned use_avx;
    short unsigned use_omp;
    int recalc;

    double x_min;
    double x_max;
    double y_min;
    double y_max;

    double zoom;
    char *args[2];
};

static struct argp_option options[] =
{
    {"output file", 'o', "FILE", 0, "Render to image, default False"},
    {"width", 'w', "WIDTH", 0, "Width of window/image, default 640"},
    {"height", 'h', "HEIGHT", 0, "Height of window/image, default 480"},
    {"iterations", 'i', "ITERATIONS", 0, "Max iterations, default 250."},
    {"dont_use_avx", 'x', 0, 0, "NOT use avx for calculations, default False"},
    {"dont_use_omp", 't', 0, 0, "NOT use openmpfor calculations, default False"},
    {"recalc", 'r', 0, 0, "Recalculate set on every frame, default False"},

    {"x_min", 'a', "VALUE", 0, "X min"},
    {"x_max", 'b', "VALUE", 0, "X max"},
    {"y_min", 'c', "VALUE", 0, "Y min"},
    {"y_max", 'd', "VALUE", 0, "Y max"},

    {"zoom", 'z', "VALUE", 0, "Zoom"},
  {0}
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
    struct arguments *arguments = state->input;
    switch (key)
    {
    case 'i':
        arguments->max_iterations = atoi(arg); 
        break;
    case 'o':
        arguments->output_file = arg; 
        break;
    case 'w':
        arguments->width = atoi(arg); 
        break;
    case 'h':
        arguments->height = atoi(arg);
        break;
    case 'x':
        arguments->use_avx = 0;
        break;
    case 't':
        arguments->use_omp = 0;
        break;
    case 'r':
        arguments->recalc = 1;
        break;

    case 'a':
        arguments->x_max = atoi(arg);
        break;
    case 'b':
        arguments->x_min = atoi(arg);
        break;
    case 'c':
        arguments->y_max= atoi(arg);
        break;
    case 'd':
        arguments->y_min= atoi(arg);
        break;
    case 'z':
        arguments->zoom= atof(arg);
        break;
    case ARGP_KEY_ARG:
        if (state->arg_num >= 2)
        /* Too many arguments. */
        argp_usage (state);
        arguments->args[state->arg_num] = arg;
      break;
    default:
        return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = {options, parse_opt, NULL, NULL};

void 
render_cl(Array *arr, SDL_Texture *texture, SDL_Renderer *renderer, int width, int height, ViewInfo *view, 
        short unsigned use_avx, short unsigned use_omp, short unsigned *needs_recalc, unsigned int max_iterations) {
    
    void (*calculate_set_func)(Array *arr, int height, int width, 
            ViewInfo view, short unsigned use_avx, unsigned max_iterations);
    
    calculate_set_func = use_omp ? calculate_set_omp : calculate_set;

    SDL_SetRenderDrawColor(renderer, 0, 0, 50, 255);
    SDL_RenderClear(renderer);
    
    if (arr->size < height * width * (int)sizeof(Color)){
        arr->pointer = realloc(arr->pointer, height * width * sizeof(Color));
        arr->size = height * width * sizeof(Color);
    }
    if (*needs_recalc){
            calculate_set_func(arr, height, width, *view, use_avx, max_iterations); 
            *needs_recalc = 0;
    }
    
    update_texture(width, height, texture, arr);

    SDL_RenderCopy(renderer, texture, NULL, NULL);
}

short unsigned
render_headless(ViewInfo view, const char* filename, 
                     int width, int height, unsigned int max_iterations,
                     short unsigned use_avx, short unsigned use_omp) {
    SDL_Surface* surface = SDL_CreateRGBSurface(0, 
            width, height, 
            32, 
            0, 0, 0, 0);

    SDL_Renderer *renderer = SDL_CreateSoftwareRenderer(surface);
    SDL_Texture* texture = SDL_CreateTexture(renderer, 
                SDL_PIXELFORMAT_RGBA8888,
                SDL_TEXTUREACCESS_STREAMING,
                width, height);

    if (!surface) {
        fprintf(stderr, "Failed to create surface: %s\n", SDL_GetError());
        return 1;
    }

    Array arr = {
        malloc(width * height * sizeof(Color)),
        width * height * sizeof(Color)
    };

    short unsigned int plug = 1; 

    render_cl(&arr, texture, renderer, width, height, &view, use_avx, use_omp, &plug, max_iterations);

    IMG_SavePNG(surface, filename);
    free(arr.pointer);
    SDL_FreeSurface(surface);
    return 0;
}


int
main(int argc, char **argv)
{
    struct arguments args = {
        NULL, 
        640, 480, 
        250, 1, 0,
        0,-2.0, 1.0, -1.5, 1.5, 1.0,
    };
    argp_parse(&argp, argc, argv, 0, 0, &args);

    if (SDL_Init(SDL_INIT_EVERYTHING) != 0){
        printf("error initializing SDL: %s\n", SDL_GetError());
        return 1;
    }
    
    ViewInfo view = {
        .x_min = args.x_min * args.zoom,
        .x_max = args.x_max * args.zoom,
        .y_min = args.y_min * args.zoom,
        .y_max = args.y_max * args.zoom,
        .zoom = args.zoom 
    };

    if (args.output_file) {
        render_headless(view, args.output_file, 
                       args.width, args.height,
                       args.max_iterations,  args.use_avx, args.use_omp
                       );
        SDL_Quit();
        return 0;
    } 

    App app = {
        SDL_CreateWindow("Mandelbrot set exploer", 
            0, 0, 
            args.width,  args.height, 
            SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE | SDL_WINDOW_OPENGL), 
        SDL_CreateRenderer(app.window, 
                -1, 
                SDL_RENDERER_ACCELERATED),
        SDL_CreateTexture(app.renderer, 
                SDL_PIXELFORMAT_RGBA8888,
                SDL_TEXTUREACCESS_STREAMING,
                args.width, args.height),
        args.height,
        args.width,
        1,
        0
            };

    
    Mouse mouse = {
        .dragging = 0,
        .start_x = 0,
        .start_y = 0,
        .original_view = view
    };

    Uint32 frame_start;

    UI ui;
    init_ui(&ui);
    
    Array screen = {
    malloc(app.win_height * app.win_width * sizeof(Color)),
    app.win_height * app.win_width * sizeof(Color)
    };
    

    int running = 1;

    while (running)
    {
        frame_start = SDL_GetTicks();
        SDL_Event event;
        while (SDL_PollEvent(&event))
        {
            switch (event.type)
            {
                case SDL_QUIT:
                    running = 0;
                    break;
                case SDL_WINDOWEVENT:
                    if(event.window.event == SDL_WINDOWEVENT_RESIZED){
                        SDL_GetWindowSize(app.window, &app.win_width, &app.win_height);
                        SDL_DestroyTexture(app.texture);

                        app.texture = SDL_CreateTexture(app.renderer, 
                        SDL_PIXELFORMAT_RGBA8888,
                        SDL_TEXTUREACCESS_STREAMING,
                        app.win_width, app.win_height);
                        app.needs_recalc = 1;
                    }
                    break;
                case SDL_KEYDOWN:
                        switch (event.key.keysym.scancode)
                        {
                            case SDL_SCANCODE_Q:
                            case SDL_SCANCODE_ESCAPE:
                                running = 0;
                                break;
                            case SDL_SCANCODE_MINUS:
                                view.x_min = view.x_min * 1.1;
                                view.x_max = view.x_max * 1.1;
                                view.y_min= view.y_min * 1.1;
                                view.y_max = view.y_max * 1.1;
                                view.zoom = view.zoom * 1.1;

                                app.needs_recalc = 1;
                                break;
                            case SDL_SCANCODE_EQUALS:
                                view.x_min = view.x_min * 0.9;
                                view.x_max = view.x_max * 0.9;
                                view.y_min= view.y_min * 0.9;
                                view.y_max = view.y_max * 0.9;
                                view.zoom = view.zoom * 0.9;

                                app.needs_recalc = 1;
                                break;
                            case SDL_SCANCODE_R:
                                view.x_min = -2.0;
                                view.x_max = 1.0;
                                view.y_min = -1.5;
                                view.y_max = 1.5;
                                view.zoom = 1.0;


                                app.needs_recalc = 1;
                                break;
                            case SDL_SCANCODE_H:
                                app.draw_ui = 1 - app.draw_ui;
                                break;
                        }
                default:
                    //returns one if recalc needed
                    if (event.type == SDL_MOUSEBUTTONDOWN || 
                        event.type == SDL_MOUSEBUTTONUP || 
                        event.type == SDL_MOUSEWHEEL || 
                        event.type == SDL_MOUSEMOTION) {
                        app.needs_recalc = handle_mouse(event, &mouse, &view);
                    }
                    app.needs_recalc = (app.needs_recalc || args.recalc);
                    break;
            }

        }
        
        //drawing happens here
        SDL_SetRenderDrawColor(app.renderer, 0xff, 0xff, 0xff, 0xff);
        SDL_RenderClear(app.renderer);

        render_cl(&screen, app.texture, app.renderer, app.win_width, app.win_height, &view, args.use_avx, args.use_omp,
                &app.needs_recalc, args.max_iterations);

        int msec = SDL_GetTicks() - frame_start;
        float fps = 0;
        if(msec > 0)
            fps = 1000.0 / (double) msec;

        if(app.draw_ui)
            render_ui(&ui, app.renderer, view, fps);

        SDL_RenderPresent(app.renderer); 
    }
        //SDL_FreeSurface(app.surfaceA);
        free(screen.pointer);

        SDL_DestroyRenderer(app.renderer);
        SDL_DestroyWindow(app.window);
        SDL_Quit();
}
